import json
import logging
from ruxit.api.base_plugin import RemoteBasePlugin
from ruxit.api.exceptions import AuthException, ConfigException
import socket
import time
import pyodbc
import ipaddress
from itoolkit import *
from itoolkit.db2.idb2call import iDB2Call

logger = logging.getLogger(__name__)
MAX_RESULTS = 100
MAX_EVENTS = 99

# TO DO:
# Add disk filter
# Add job status

class CustomIBMiPluginRemote(RemoteBasePlugin):
    last_collect_creation = 0
    createdPerfData = False
    metricHolder = {} # To store deltas
    collection_library = ''
    host = None
    cm = {} # Collection Member
    
    events = 0
    previousEvents = []
    firstExecution = True
    
    def query(self, **kwargs):
        self.config = kwargs['config']
        
        if self.firstExecution or self.config['debug']:
            logging.basicConfig(level=logging.DEBUG)
            logger.setLevel(logging.DEBUG)
        else:
            logging.basicConfig(level=logging.INFO)
            logger.setLevel(logging.INFO)
            
        # Config GUI vars
        host = self.config['host'].strip()
        group = self.config['group'].strip() if self.config['group'].strip() != '' else 'Default group'
        username = self.config['auth_user'].strip()
        password = self.config['auth_password']
        self.jobsFilter = self.config['jobs'].strip()
        self.messageQueuesFilter = self.config['message_queues'].strip()
        self.jobQueuesFilter = self.config['job_queues'].strip()
        cursor = None
        self.conn = None
        sql = ''
        self.jobsFilterSql = ''
        self.messageQueuesFilterSql = ''
        self.jobQueuesFilterSql = ''
        pfrDataColExists = False
        self.perf_data_source = self.config['perf_data_source']
        self.version = None
        self.max_disks = 100 if 'max_disks' not in self.config else int(self.config['max_disks'])
        use_ssl = True if 'ssl' in self.config and str(self.config['ssl']).lower() == 'true' else False
        
        logger.info(f'Polling to IBM i {host} started')
        
        if host == '':
            raise ConfigException('Hostname or IP cannot be empty. You must provide this to connect to an IBM i host')

        if username == '' or password == '':
            raise ConfigException('You must provide a username and password to connect to this IBM i host')
            
        if self.jobsFilter != '':
            self.jobsFilter = self.jobsFilter.split(',',MAX_RESULTS)
            if len(self.jobsFilter) == MAX_RESULTS+1:
                self.jobsFilter.pop()
                
            if len(self.jobsFilter) > 0:
                self.jobsFilter = [f"'{item.strip().upper()}'" for item in self.jobsFilter]
                self.jobsFilterSql = f" AND TRIM(J.jbname) IN ({','.join(self.jobsFilter)}) "
                
        if self.messageQueuesFilter != '':
            if self.messageQueuesFilter != '*':
                self.messageQueuesFilter = self.messageQueuesFilter.split(',',MAX_RESULTS)
                if len(self.messageQueuesFilter) == MAX_RESULTS+1:
                    self.messageQueuesFilter.pop()
                    
                if len(self.messageQueuesFilter) > 0:
                    self.messageQueuesFilter = [f"'{item.strip().upper()}'" for item in self.messageQueuesFilter]
                    self.messageQueuesFilterSql = f"WHERE MESSAGE_QUEUE_NAME IN({','.join(self.messageQueuesFilter)})"
                    
        if self.jobQueuesFilter != '':
            if self.jobQueuesFilter != '*':
                self.jobQueuesFilter = self.jobQueuesFilter.split(',',MAX_RESULTS)
                if len(self.jobQueuesFilter) == MAX_RESULTS+1:
                    self.jobQueuesFilter.pop()
                    
                if len(self.jobQueuesFilter) > 0:
                    self.jobQueuesFilter = [f"'{item.strip().upper()}'" for item in self.jobQueuesFilter]
                    self.jobQueuesFilterSql = f"WHERE JOB_QUEUE_NAME IN({','.join(self.jobQueuesFilter)})"
        
        hostIP = None
            
        # Check to see if the host/ip is valid
        try:
            hostIP = socket.gethostbyname(host)
        except socket.error as e:
            raise ConfigException(f'IP could not be determined: {host}')
        
        self.host = host
        
        if host not in self.cm:
            self.cm[host] = {}
        
        g1 = self.topology_builder.create_group(f'IBM i - {group}', f'IBM i - {group}')
        node = g1.create_element(f'IBM i - {group} - {host}', f'IBM i - {group} - {host}')
        node.report_property('Plugin version', kwargs['json_config']['version'])
        
        self.itool = iToolKit()
        try:
            conn_string = "DRIVER={IBM i Access ODBC Driver};SYSTEM="+hostIP+";UID="+username+";PWD={"+password+"};RetryParameterBindingOnError=false"
            if use_ssl:
                conn_string += ";SSL=1"
            self.conn = pyodbc.connect(conn_string, timeout=5)
            cursor = self.conn.cursor()
            
            if hostIP:
                node.add_endpoint(hostIP)
            
            # Get the OS version
            cursor.execute("SELECT OS_VERSION || '.' || OS_RELEASE AS OS_VERSION FROM SYSIBMADM.ENV_SYS_INFO")
            row = cursor.fetchone()
            if row:
                try:
                    self.version = float(row.OS_VERSION)
                except ValueError as e:
                    logger.warn(f'Could not convert OS version {row.OS_VERSION}to number. Assuming <7.2')
                    self.version = 7.1
            
            # Get TR level
            cursor.execute("SELECT PTF_GROUP_LEVEL AS TR_LEVEL FROM QSYS2.GROUP_PTF_INFO WHERE PTF_GROUP_DESCRIPTION = 'TECHNOLOGY REFRESH' AND PTF_GROUP_STATUS = 'INSTALLED' ORDER BY PTF_GROUP_LEVEL DESC FETCH FIRST 1 ROWS ONLY")
            row = cursor.fetchone()
            self.tr_level = 0
            if row:
                self.tr_level = int(row.TR_LEVEL)

            logger.info(f'TR level = {self.tr_level}')
            
            # Begin querying data
            if 'Catalog' not in self.perf_data_source:
                # Run Performance Collection queries
                self.cm[host]['member'] = self.getLastPerformanceMember(cursor)
              
                if (self.cm[self.host]['member']):
                    try:
                        pfrDataColExists = self.checkPerfDataColExists(cursor)
                    except Exception as e:
                        logger.exception(f'IBM i - Could not check if pfrDataColExists, assuming it does not: {e}')
                        self.report_custom_info_event(node, 'Could not check if Performance collection member exists. Choose a different data source on configuration UI or consult with an IBM i admin', 'Could not check if Perf Collection member exists')
                    
                    if pfrDataColExists:
                    
                        self.initialize_queries()
                        
                        set_aliases = []
                        for query in self.queries:
                            try:
                                if 'aliasMap' in query:
                                    for alias in query['aliasMap']:
                                        if alias["alias"] not in set_aliases:
                                            logger.debug(f'Executing alias query: CREATE OR REPLACE ALIAS {alias["alias"]} FOR {alias["fileTable"]} ({self.cm[self.host]["member"]})')
                                            cursor.execute(f'CREATE OR REPLACE ALIAS {alias["alias"]} FOR {alias["fileTable"]} ({self.cm[self.host]["member"]})')
                                            set_aliases.append(alias["alias"])
                                # Only run the queries if there was no problem creating alias and after each alias.
                                logger.debug(f'Executing APM query: {query["sql"]}')
                                cursor.execute(query['sql'])
                                rows = cursor.fetchall()
                                self.iterateResults(rows, query, node)
                            except pyodbc.Error as e2:
                                sqlState = e2.args[0]
                                errMsg = e2.args[1]
                                if 'aliasMap' in query:
                                    for alias in query['aliasMap']:
                                        if alias["alias"] not in set_aliases:
                                            logger.error(f'An error occurred creating alias. Perhaps the file table does not exist in Performance Collection. {errMsg}. Query: CREATE OR REPLACE ALIAS {alias["alias"]} FOR {alias["fileTable"]} ({self.cm[self.host]["member"]})')
                                            node.report_custom_info_event(f'Could not get performance data from host {host}. Please ensure that you have performance collection services running and collecting data for category {query["category"]}', 'Error collecting performance data', properties={'Category': query['category'], 'Source': self.perf_data_source, 'Library': self.collection_library})
                                else:
                                    logger.error(f'An error occurred executing query. {errMsg}. Query: {query["sql"]}')
                                    node.report_custom_info_event(f'Error occurred running query on {host}. {errMsg}', 'Error executing query', properties={'Source': self.perf_data_source, 'Library': self.collection_library})
                    else:
                        self.createPerfDataCol(node)
                elif 'Overwrite' in self.perf_data_source: # It couldn't get the last member of Performance collection. Probably it does not exist
                    self.createPerfDataCol(node)
                else: # It is expecting an existing Performance Collection. Error out
                    logger.error(f'No existing member of Performance Collection was found. Query to QAPMSYSTEM returned no rows. Ensure existing Performance Collection profile has all categories collecting data (SYSLVL, DISK, JOBMI, JOBOS, POOL, SYSCPU)')
                    if self.firstExecution: # If it's the first time running, let's show a ConfigException
                        raise ConfigException('Unable to get latest partition member of EXISTING Performance Collection. Ensure necessary categories are collecting data in your existing Performance collection profile (SYSLVL, DISK, JOBMI, JOBOS, POOL, SYSCPU)')
                    else:
                        self.report_error_event(f'No existing member of Performance Collection was found. Ensure all necessary categories are configured on your existing Performance Collection profile (SYSLVL, DISK, JOBMI, JOBOS, POOL, SYSCPU)', 'Error collecting data from IBM i server')
            
            else:
                self.initialize_queries()
                
                # Run Catalog table queries instead
                for query in self.queries:
                    logger.debug(f"Catalog query: {query['sql']}")
                    cursor.execute(query['sql'])
                    #logger.info([column[0] for column in cursor.description])
                    rows = cursor.fetchall()
                    self.iterateResults(rows, query, node)

        except pyodbc.Error as e:
            sqlState = e.args[0]
            errMsg = e.args[1]
            logger.error(errMsg)
            if ('member' in self.cm[self.host] and self.cm[self.host]['member'] and self.cm[self.host]['member'] in errMsg) and 'not found' in errMsg:
                logger.warn('Did not find collection name in errMsg. Creating PerfDataCol')
                self.createPerfDataCol(node)
            else:
                if sqlState == '28000': # Bad Authentication
                    node.report_error_event(f'Wrong username {self.config["auth_user"]}, password or account disabled', 'Authentication error')
                    logger.error(f'Authentication error: {e}')
                    raise AuthException(f'{errMsg}')
                elif sqlState == '08S01':
                    node.report_error_event(f'Cannot communicate with {host}. Network connectivity error.', f'Cannot communicate with {host}')
                    logger.error(f'Cannot communicate with {host}. Network connectivity error. {errMsg}')
                    raise ConfigException(f'Cannot communicate with {host}. Network connectivity error. {errMsg}')
                else:
                    logger.error(f'Failed to query IBM i on {host}. {sql}. {errMsg}')
                    raise ConfigException(f'Could not query IBM i host {host}. Error: {errMsg}')
        finally:
            if cursor:
                cursor.close()
            if self.conn:
                self.conn.close()
                logger.info(f'Closed connection')
        
        # After executing once
        self.firstExecution = False 
        
        logger.info(f'Finished polling to IBM i {host}')

    def runQuery(self, query, node):
        cursor.execute(query['sql'])
        rows = cursor.fetchall()
        self.iterateResults(rows, query, node)
    
    def createPerfDataCol(self, node):
        if 'Overwrite' in self.perf_data_source:
            if self.createdPerfData == False:
                transport = iDB2Call(self.conn)
                try:
                    self.itool.add(iCmd('cfgpfrcol', 'CFGPFRCOL CYCTIME(000000) CYCITV(24) RETPERIOD(1 *DAYS) ENBSYSMON(*YES) SYSMONRET(1) CRTPFRSUM(*SYSMON) SYSMONCGY((*SYSLVL 1) (*DISK 1) (*JOBMI 1) (*JOBOS 1) (*POOL 1) (*SYSCPU 1))'))
                    self.itool.add(iCmd('strpfrcol', 'STRPFRCOL COLPRF(*CUSTOM) CYCCOL(*YES)'))
                    
                    # xmlservice
                    self.itool.call(transport)
                    logger.info('Sent command to create PerfData collection. Now to wait...')
                    self.last_collect_creation = time.time()
                    self.createdPerfData = True
                except Exception as e2:
                    logger.error(f'Could not create Performance Data Collection. {e2}')
                    self.report_custom_info_event(node, f'Could not create Performance Data Collection on IBM i. Cannot gather performance data.', 'Error creating PerfData collection')

            # If we attempted to create Performance data collection before and it's been more than 5 minutes
            # and still we can't access performance data. Give up.
            elif self.createdPerfData == True and time.time()-self.last_collect_creation > 300:
                logger.error(f'Could not create Performance Data Collection.')
                self.report_custom_info_event(node, f'Could not create Performance Data Collection on IBM i. Cannot gather performance data.', 'Error creating PerfData collection')
            
    def iterateResults(self, rows, metrics, node):
        for row in rows:
            logger.debug("row: " + str(row))
            if 'metricMap' in metrics:
                for map in metrics['metricMap']:
                    
                    if 'sqlFieldAlias' in map:
                        map['sqlField'] = map['sqlFieldAlias']
                    
                    val = getattr(row,map['sqlField'])
                    if 'owner' in map and map['owner'] == 'property':
                        displayField = map['displayField']
                        if 'Memory Pool' in displayField and 'splitBy' in metrics and getattr(row,metrics['splitBy']).strip() not in ['*BASE', '*MACHINE', '*INTERACT', '*SPOOL']:
                            continue
                        if 'format' in map:
                            logger.debug(f'formatting {val}')
                            val = "{:,}".format(int(val))
                        if 'unit' in map:
                            val = f'{val} {map["unit"]}'
                        if 'splitBy' in metrics:
                            displayField = displayField.replace(f'[{metrics["splitBy"]}]', getattr(row,metrics['splitBy']).strip())
                        node.report_property(displayField, str(val))
                    if 'metricName' in map:
                        if 'splitBy' in metrics:
                            metricDimension = {metrics['splitName']:getattr(row,metrics['splitBy'])}
                            if 'delta' in map:
                                key = self.host+'_'+list(metricDimension.values())[0]+'_'+map['metricName']
                                if key not in self.metricHolder:
                                    self.metricHolder[key] = val
                                    continue
                                else:
                                    logger.debug(f'Calculating delta for ({self.host}) {key}: {self.metricHolder[key]} - {val} = {abs(self.metricHolder[key]-val)}')
                                    delta = abs(self.metricHolder[key]-val)
                                    self.metricHolder[key] = val
                                    val = delta
                            if val is not None:
                                node.absolute(key=map['metricName'], value=float(val), dimensions=metricDimension)
                                logger.debug(f'Setting value for dimension {metricDimension} in {map["metricName"]} = {val}')
                        else:
                            val = getattr(row, map['sqlField'])
                            if val is not None:
                                node.absolute(key=map['metricName'], value=float(val))
                            
    def getLastPerformanceMember(self, cursor):
        #sql = "SELECT TABLE_SCHEMA,TABLE_PARTITION FROM QSYS2.SysPartitionStat Where Table_Name = 'QAPMSYSTEM' Order By Partition_Number Desc"
        if self.createdPerfData or self.perf_data_source.startswith('SYSMON'):
            member_prefix = 'R'
        else:
            member_prefix = 'Q'
        sql = f"SELECT TABLE_SCHEMA,TABLE_PARTITION FROM QSYS2.SysPartitionStat Where Table_Name = 'QAPMSYSTEM' AND TABLE_SCHEMA in ('QPFRDATA', 'QMPGDATA') AND TABLE_PARTITION LIKE '{member_prefix}%' Order By Partition_Number Desc"
        cursor.execute(sql)
        #logger.info([column[0] for column in cursor.description])
        row = cursor.fetchone()
        logger.debug(f'rowcount: {cursor.rowcount}. resultset: {row}')
        
        if row:
            #activeMember = row.TABLE_PARTITION
            self.collection_library = row.TABLE_SCHEMA
        
            #if self.createdPerfData or ('member' in self.cm[self.host] and self.cm[self.host]['member'].startswith('R')):
            '''
            if self.createdPerfData or self.perf_data_source.startswith('SYSMON'):
                self.cm[self.host]['member'] = f'R{activeMember[1:]}'
            else:
                self.cm[self.host]['member'] = activeMember
            '''
            self.cm[self.host]['member'] = str(row.TABLE_PARTITION).strip()
            logger.debug(f'ActiveMember: {self.cm[self.host]["member"]}')
            
            return self.cm[self.host]['member']
        return None
    
    # There could be several reasons why a Perf Data collector is not running:
    # - First time running
    # - It is running but at a different interval, not the 1min interval we need
    # - The file member does not exist and must be created/recreated
    # - Collector was stopped for any reason: FileSystem is full, manual stop, etc.
    def checkPerfDataColExists(self, cursor):
        # Check interval
        # If for any reason the following fails, collector is not running
        cursor.execute(f'CREATE OR REPLACE ALIAS QTEMP.PERFSYSTEM FOR {self.collection_library}.QAPMSYSTEM ({self.cm[self.host]["member"]})')
        sql = 'SELECT INTSEC FROM QTEMP.PERFSYSTEM WHERE INTNUM = (SELECT MAX(INTNUM) FROM QTEMP.PERFSYSTEM) GROUP BY INTSEC'
        cursor.execute(sql)
        row = cursor.fetchone()
        if row: # Found something
            logger.debug(f'INTSEC resultset: {row} in member {self.cm[self.host]["member"]}')
            if 'Overwrite' in self.perf_data_source and int(row.INTSEC) != 60:
                return False
                
            # If we get to this point then a record was found so there is performance collection
            # but is either DEFAULT or an existing SYSMON collection
            return True
   
        # Found nothing or interval is not one we can use
        return False
        
    def report_custom_info_event(self, e1, message, title):
        if self.events < MAX_EVENTS:
            if title + message not in self.previousEvents:
                self.events += 1
                e1.report_custom_info_event(message, title)
                self.previousEvents.append(title + message)
                
    def initialize_queries(self):
        apmDiskQuery = f"select TRIM(CAST(DSDRN AS CHAR (10) CCSID 1399))||'/'||TRIM(CAST(DSARM AS CHAR (10) CCSID 1399))||TRIM(CAST(DMFLAG AS CHAR (10) CCSID 1399)) AS DISKUNITNAME, \
                                    avg(CASE WHEN DSSMPL=0 THEN NULL ELSE 100*(DOUBLE(DSSMPL-DSNBSY)/DSSMPL) END) AS DISKBUSY, \
                                    sum(case when binary(land(DSCAT,binary(x'01')))=binary(x'01') then DOUBLE(DSDWOP+DSDROP) else DOUBLE(DSRDS+DSWRTS) end) AS DISKTOTALOP, \
                                    sum(case when binary(land(DSCAT,binary(x'01')))=binary(x'01') then DOUBLE(DSDWOP + DSDROP)/INTSEC else DOUBLE(DSRDS+DSWRTS)/INTSEC end) AS DISKTOTALOPRATE, \
                                    SUM(DOUBLE(DSCAP/1024)/1024) AS DISKCAPACITY, \
                                    SUM(((DOUBLE(DSCAP)-DOUBLE(DSAVL))/1024)/1024) AS DISKUSED, \
                                    SUM(DOUBLE(DSRDS)) / AVG(INTSEC) AS READOPRATE, \
                                    SUM(DOUBLE(DSWRTS)) / AVG(INTSEC) AS WRITEOPRATE, \
                                    sum(double(DSBLKR) * 520 / 1024 ) as TOTALKBR, \
                                    sum(double(DSBLKW) * 520 / 1024 ) as TOTALKBW, \
                                    sum(double(DSBLKR) * 520 / 1024 )/AVG(INTSEC) as RATEKBR, \
                                    sum(double(DSBLKW) * 520 / 1024 )/AVG(INTSEC) as RATEKBW, \
                                    sum(case when binary(land(DSCAT,binary(x'01')))=binary(x'01') then DOUBLE(DSDWOP) else DOUBLE(DSWRTS) end) as DISKWRITES, \
                                    sum(case when binary(land(DSCAT,binary(x'01')))=binary(x'01') then DOUBLE(DSDROP) else DOUBLE(DSRDS) end) AS DISKREADS, \
                                    sum(case when binary(land(DSCAT,binary(x'01')))=binary(x'01') then case when DSDROP=0 then NULL else 100*(DOUBLE(DSDCRH)/DOUBLE(DSDROP)) END else CASE WHEN DSRDS=0 THEN NULL ELSE 100*(DOUBLE(DSCCRH)/DOUBLE(DSRDS)) END end) AS READCACHERATIO, \
                                    sum(case when DSWRTS=0 then NULL else 100*(100-DOUBLE(DSCCFW))/(DOUBLE(DSWRTS)) end) AS WRITECACHERATIO, \
                                    SUM(DOUBLE(DSSRVT+DSWT)/DOUBLE(DSWRTS+DSRDS)) AS DISKRESPTIME, \
                                    SUM(DOUBLE(DSSRVT)/DOUBLE(DSWRTS+DSRDS)) AS DISKAVGSRVTIME, \
                                    SUM(DOUBLE(DSWT)/DOUBLE(DSWRTS+DSRDS)) AS DISKAVGWAITTIME \
                                    from QTEMP.PERFDISK \
                                        WHERE INTNUM = (SELECT MAX(INTNUM) FROM QTEMP.PERFDISK) \
                                        GROUP BY TRIM(CAST(DSDRN AS CHAR (10) CCSID 1399))||'/'||TRIM(CAST(DSARM AS CHAR (10) CCSID 1399))||TRIM(CAST(DMFLAG AS CHAR (10) CCSID 1399)) FETCH FIRST {self.max_disks} ROWS ONLY"
        '''
        apmCpuQuery = 'SELECT (CASE WHEN SYSCTA=0 THEN NULL ELSE SYSPTU/DOUBLE(SYSCTA) * 100 END) AS PCTTOTCPU, \
                                            (case when SYIFUS+SYIFTA=0 then NULL else SYIFUS/DOUBLE((SYIFUS+SYIFTA)) * 100 end) AS PCTINTCPU, \
                                            (case when SYSPTU=0 then NULL else DOUBLE(SYSSPTU/SYSPTU) * 100 end) AS PROC_SPEED \
                                        FROM QTEMP.PERFSYSTEM WHERE INTNUM = (SELECT MAX(INTNUM) FROM QTEMP.PERFSYSTEM) \
                                        ORDER BY INTNUM DESC'
        '''
        apmCpuQuery = 'SELECT DOUBLE(SMSCPUCPU) AS PCTTOTCPU, DOUBLE(SMSPRCSPDP) AS PROC_SPEED \
                                        FROM QTEMP.PERFSMSYS WHERE INTNUM = (SELECT MAX(INTNUM) FROM QTEMP.PERFSMSYS) \
                                        ORDER BY INTNUM DESC'
        
        apmIntCpuQuery = 'SELECT (case when SYIFUS+SYIFTA=0 then NULL else SYIFUS/DOUBLE((SYIFUS+SYIFTA)) * 100 end) AS PCTINTCPU \
                                        FROM QTEMP.PERFSYSTEM WHERE INTNUM = (SELECT MAX(INTNUM) FROM QTEMP.PERFSYSTEM) \
                                        ORDER BY INTNUM DESC'
                                        
        apmVirtualCpuQuery = "SELECT TRIM(CAST(VPID AS CHAR (10) CCSID 1399)) AS VCPUID, \
                                            (CASE WHEN VPCPUAC=0 THEN NULL ELSE 100*VPCPUU/DOUBLE(VPCPUAC) END) AS VCPUPCT \
                                        from QTEMP.PERFVCPU WHERE VPACTF='1' AND INTNUM = (SELECT MAX(INTNUM) FROM QTEMP.PERFVCPU)"
        apmJobQuery = f"SELECT * FROM (SELECT CASE WHEN TRIM(CAST(L.jbssys AS CHAR (10) CCSID 1399)) IS NOT NULL THEN (TRIM(CAST(L.jbuser AS CHAR (10) CCSID 1399))||'/'||TRIM(CAST(L.jbssys AS CHAR (10) CCSID 1399))||'/'||TRIM(CAST(J.jbname AS CHAR (10) CCSID 1399))) ELSE (TRIM(CAST(L.jbuser AS CHAR (10) CCSID 1399))||'/'||TRIM(CAST(J.jbname AS CHAR (10) CCSID 1399))) END AS JOBNAME, \
                                                AVG(CASE WHEN S.syscta=0 THEN NULL ELSE CAST (J.jbcpu / S.syscta * 100 AS DECIMAL (15, 2)) END) AS JOBCPU, \
                                                SUM(J.jbdbr + J.jbndb + J.jbadbr + J.JBANDR + J.jbwrt) AS JOBIO, \
                                                SUM(J.JBTFLT) AS JOBPAGEFAULTS, \
                                                SUM(J.JBSLTCNT) AS JOBSHORTLIVED, \
                                                SUM(L.JBSPD) AS JOBSUSPENDTIME \
                                        FROM QTEMP.PERFJOBMI J, QTEMP.PERFSYSL S, QTEMP.PERFJOBL L \
                                        WHERE J.JBNAME=L.JBNAME AND J.INTNUM=S.INTNUM AND S.INTNUM=L.INTNUM AND J.INTNUM = (SELECT MAX(INTNUM) FROM QTEMP.PERFJOBMI) {self.jobsFilterSql} \
                                        GROUP BY CASE WHEN TRIM(CAST(L.jbssys AS CHAR (10) CCSID 1399)) IS NOT NULL THEN (TRIM(CAST(L.jbuser AS CHAR (10) CCSID 1399))||'/'||TRIM(CAST(L.jbssys AS CHAR (10) CCSID 1399))||'/'||TRIM(CAST(J.jbname AS CHAR (10) CCSID 1399))) ELSE (TRIM(CAST(L.jbuser AS CHAR (10) CCSID 1399))||'/'||TRIM(CAST(J.jbname AS CHAR (10) CCSID 1399))) END) A \
                                        ORDER BY A.JOBCPU DESC FETCH FIRST {MAX_RESULTS} ROWS ONLY"
                                        
        apmMemPoolQuery = f"SELECT \
                            TRIM(CAST(PONBR AS CHAR (10) CCSID 1399)) AS POOLNUMBER, \
                            TRIM(CAST(QSYS2.MEMORY_POOL_INFO.POOL_NAME AS CHAR (10) CCSID 1399)) AS POOLNAME, \
                            CASE \
                              WHEN length(TRIM(CAST(PONBR AS CHAR (10) CCSID 1399)))=3 \
                              THEN substr(TRIM(CAST(PONBR AS CHAR (10) CCSID 1399)),2,3) \
                              ELSE TRIM(CAST(PONBR AS CHAR (10) CCSID 1399)) \
                            END AS POOLNUMBER_SHORT, \
                            DOUBLE(POSIZ)/DOUBLE(1028) AS POOLSIZE, \
                            (DOUBLE(PODBF)+DOUBLE(PONDBF)) AS POOLFAULT, \
                            (DOUBLE(PODBF)+DOUBLE(PONDBF))/INTSEC AS POOLFAULTRATE, \
                            DOUBLE(PODBF) AS POOLDBFAULT, \
                            DOUBLE(PODBF)/INTSEC AS POOLDBFAULT_RATE, \
                            DOUBLE(PONDBF) AS POOLNONDBFAULT, \
                            DOUBLE(PONDBF)/INTSEC AS POOLNONDBF_RATE \
                        FROM \
                            QTEMP.PERFPOOLB T1, QSYS2.MEMORY_POOL_INFO\
                        WHERE \
                            QSYS2.MEMORY_POOL_INFO.SYSTEM_POOL_ID=CAST(T1.PONBR AS SMALLINT) AND \
                            INTNUM = (SELECT MAX(INTNUM) FROM QTEMP.PERFPOOLB) \
                        ORDER BY \
                            PONBR ASC"
                            
        apmSharedCpuPoolQuery = "SELECT \
                                AVG(DOUBLE(SMSVSPP)) AS SHARED_PROC_POOL_UTIL_VIRT, \
                                AVG(DOUBLE(SMSPSPP)) AS SHARED_PROC_POOL_UTIL_PHYS \
                            FROM QTEMP.PERFSMSYS WHERE INTNUM = (SELECT MAX(INTNUM) FROM QTEMP.PERFSMSYS)"
        
        self.queries = []
        
        if 'Catalog' not in self.config['perf_data_source']:
            self.queries.append({'name': 'poolStatus', 'category': 'Memory pool (*POOL)', 'metricMap': [{'displayField': '[POOL_NAME] Memory Pool size', 'sqlField': 'CURRENT_SIZE', 'owner': 'property', 'format':'true', 'unit': 'MB'}, {'sqlField': 'CURRENT_THREADS', 'metricName':'Pool.Threads'}], 'aliasMap':[], 'sqlTable': 'QSYS2.MEMORY_POOL_INFO', 'sql': "SELECT TRIM(CAST(POOL_NAME AS CHAR (10) CCSID 1399)) AS POOL_NAME, SYSTEM_POOL_ID, CURRENT_SIZE, CURRENT_THREADS, ELAPSED_TOTAL_FAULTS FROM QSYS2.MEMORY_POOL_INFO", 'owner':'Memory', 'splitName': 'Memory', 'splitBy': 'POOL_NAME', 'metricType': 'absolute'})
            self.queries.append({'name': 'apmCpu', 'category': 'System CPU', 'metricMap': [{'sqlField': 'PROC_SPEED', 'metricName': 'System.CpuSpeed'}, {'sqlField': 'PCTTOTCPU', 'metricName':'System.Cpu'}], 'aliasMap':[{'alias': 'QTEMP.PERFSMSYS', 'fileTable': f'{self.collection_library}.QAPMSMSYS'}], 'sql': apmCpuQuery, 'owner':'CPU', 'metricType': 'absolute'})
            self.queries.append({'name': 'apmIntCpu', 'category': 'System CPU', 'metricMap': [{'sqlField': 'PCTINTCPU', 'metricName':'System.InteractiveCpu'}], 'aliasMap':[{'alias': 'QTEMP.PERFSYSTEM', 'fileTable': f'{self.collection_library}.QAPMSYSTEM'}], 'sql': apmIntCpuQuery, 'owner':'CPU', 'metricType': 'absolute'})
            self.queries.append({'name': 'apmVirtualCpu', 'category': 'System CPU', 'metricMap': [{'sqlField': 'VCPUPCT', 'metricName':'CPU.Utilization'}], 'aliasMap':[{'alias': 'QTEMP.PERFVCPU', 'fileTable': f'{self.collection_library}.QAPMSYSVP'}], 'splitName': 'CPU', 'splitBy': 'VCPUID', 'sql': apmVirtualCpuQuery, 'owner':'CPU', 'metricType': 'absolute'})
            self.queries.append({'name': 'apmDisk', 'category': 'Disk storage', 'aliasMap':[{'alias': 'QTEMP.PERFDISK', 'fileTable': f'{self.collection_library}.QAPMDISK'}], 'metricMap': [{'sqlField': 'DISKUSED', 'metricName': 'Disk.Used'}, {'sqlField': 'DISKCAPACITY', 'metricName': 'Disk.Capacity'}, {'sqlField': 'WRITEOPRATE', 'metricName': 'Disk.WriteOpRate'}, {'sqlField': 'DISKTOTALOP', 'metricName': 'Disk.TotalOp'}, {'sqlField': 'READOPRATE', 'metricName': 'Disk.ReadOpRate'}, {'sqlField': 'DISKTOTALOPRATE', 'metricName': 'Disk.TotalOpRate'}, {'sqlField': 'DISKBUSY', 'metricName': 'Disk.Busy'}, {'sqlField': 'DISKREADS', 'metricName': 'Disk.Reads'}, {'sqlField': 'DISKWRITES', 'metricName': 'Disk.Writes'}, {'sqlField': 'TOTALKBR', 'metricName': 'Disk.TotalKBReads'}, {'sqlField': 'TOTALKBW', 'metricName': 'Disk.TotalKBWrites'}, {'sqlField': 'RATEKBR', 'metricName': 'Disk.RateKBReads'}, {'sqlField': 'RATEKBW', 'metricName': 'Disk.RateKBWrites'}, {'sqlField': 'READCACHERATIO', 'metricName': 'Disk.ReadCacheRatio'}, {'sqlField': 'WRITECACHERATIO', 'metricName': 'Disk.WriteCacheRatio'}, {'sqlField': 'DISKRESPTIME', 'metricName': 'Disk.ResponseTime'}, {'sqlField': 'DISKAVGSRVTIME', 'metricName': 'Disk.AvgServiceTime'}, {'sqlField': 'DISKAVGWAITTIME', 'metricName': 'Disk.AvgWaitTime'}], 'splitName': 'Disk', 'splitBy': 'DISKUNITNAME', 'sql': apmDiskQuery, 'owner':'Disk', 'metricType': 'absolute'})
            self.queries.append({'name': 'apmPools', 'category': 'Memory pool', 'metricMap': [{'sqlField': 'POOLFAULT', 'metricName': 'Pool.Faults'}, {'sqlField': 'POOLFAULTRATE', 'metricName': 'Pool.FaultRate'}, {'sqlField': 'POOLDBFAULT', 'metricName': 'Pool.DBFault'}, {'sqlField': 'POOLDBFAULT_RATE', 'metricName': 'Pool.DBFaultRate'}, {'sqlField': 'POOLNONDBFAULT', 'metricName': 'Pool.NonDBFault'}, {'sqlField': 'POOLNONDBF_RATE', 'metricName': 'Pool.NonDBFaultRate'}], 'aliasMap': [{'alias': 'QTEMP.PERFPOOLB', 'fileTable': f'{self.collection_library}.QAPMPOOLB'},{'alias': 'QTEMP.PERFJOBL', 'fileTable': f'{self.collection_library}.QAPMJOBL'}], 'sql': apmMemPoolQuery, 'owner':'Memory', 'splitName': 'Memory', 'splitBy': 'POOLNAME', 'metricType': 'absolute'})
            self.queries.append({'name': 'apmSharedCpuPool', 'category': 'System-level data', 'metricMap': [{'sqlField': 'SHARED_PROC_POOL_UTIL_VIRT', 'metricName': 'System.SharedCpuPoolVirtual'}, {'sqlField': 'SHARED_PROC_POOL_UTIL_PHYS', 'metricName': 'System.SharedCpuPoolPhys'}], 'aliasMap': [{'alias': 'QTEMP.PERFSMSYS', 'fileTable': f'{self.collection_library}.QAPMSMSYS'}], 'sql': apmSharedCpuPoolQuery, 'owner':'CPU', 'splitName': 'Memory', 'metricType': 'absolute'})
            if self.jobsFilterSql != '':
                self.queries.append({'name': 'apmJobs', 'category': 'Jobs (MI tasks and Operating system)', 'metricMap': [{'sqlField': 'JOBCPU', 'metricName':'Job.Cpu'}, {'sqlField': 'JOBIO', 'metricName': 'Job.IO'}, {'sqlField': 'JOBSHORTLIVED', 'metricName': 'Job.ShortLived'}, {'sqlField': 'JOBPAGEFAULTS', 'metricName': 'Job.PageFaults'}, {'sqlField': 'JOBSUSPENDTIME', 'metricName': 'Job.SuspendTime'}], 'aliasMap': [{'alias': 'QTEMP.PERFJOBMI', 'fileTable': f'{self.collection_library}.QAPMJOBMI'}, {'alias': 'QTEMP.PERFSYSL', 'fileTable': f'{self.collection_library}.QAPMSYSL'}], 'sql': apmJobQuery, 'owner':'Jobs', 'splitName': 'Job', 'splitBy': 'JOBNAME', 'metricType': 'absolute'})
        else:
            self.queries.append({'name': 'systemStatus1', 'metricMap': [{'sqlField': 'INTERACTIVE_JOBS_IN_SYSTEM', 'metricName':'System.InteractiveCpu', 'owner': 'System'}, {'sqlField': 'AVERAGE_CPU_UTILIZATION', 'metricName':'System.Cpu', 'owner': 'System'}], 'sql': 'SELECT INTERACTIVE_JOBS_IN_SYSTEM,AVERAGE_CPU_UTILIZATION FROM QSYS2.SYSTEM_STATUS_INFO', 'metricType': 'absolute'})
            self.queries.append({'name': 'poolStatus', 'metricMap': [{'displayField': '[POOL_NAME] Memory Pool size', 'sqlField': 'CURRENT_SIZE', 'owner': 'property', 'format':'true', 'unit': 'MB'}, {'sqlField': 'CURRENT_THREADS', 'metricName':'Pool.Threads', 'owner': 'Memory'}, {'sqlField': 'ELAPSED_TOTAL_FAULTS', 'metricName':'Pool.FaultRate', 'owner': 'Memory'}, {'sqlField': 'POOL_NAME', 'owner': 'Memory'}], 'sql': "SELECT CURRENT_SIZE,CURRENT_THREADS,ELAPSED_TOTAL_FAULTS,POOL_NAME FROM QSYS2.MEMORY_POOL_INFO WHERE TRIM(POOL_NAME) IN ('*BASE', '*MACHINE', '*INTERACT', '*SPOOL')", 'splitName': 'Memory', 'splitBy': 'POOL_NAME', 'metricType': 'absolute'})
            if self.jobsFilterSql != '':
                self.queries.append({'name': 'jobStatus', 'metricMap': [{'sqlField': "CASE WHEN SUBSYSTEM IS NOT NULL THEN SUBSTR(JOB_NAME,8,POSSTR(SUBSTR(JOB_NAME,8),'/')-1)||'/'||SUBSYSTEM||'/'||SUBSTR(SUBSTR(JOB_NAME,8),POSSTR(SUBSTR(JOB_NAME,8),'/')+1) ELSE RIGHT(JOB_NAME, LENGTH(JOB_NAME)-LOCATE('/',JOB_NAME)) END", 'sqlFieldAlias': 'JOB_NAME', 'owner': 'Job'}, {'sqlField': 'SUM(DOUBLE(CPU_TIME))', 'sqlFieldAlias': 'CPUTIME', 'metricName':'Job.CpuTime', 'owner': 'Job', 'delta': 'true'}, {'sqlField': 'SUM(DOUBLE(TOTAL_DISK_IO_COUNT))', 'delta': 'true', 'metricName':'Job.IO', 'owner': 'Job', 'sqlFieldAlias': 'TOTAL_DISK_IO_COUNT'}], 'sql': f"SELECT CASE WHEN SUBSYSTEM IS NOT NULL THEN SUBSTR(JOB_NAME,8,POSSTR(SUBSTR(JOB_NAME,8),'/')-1)||'/'||SUBSYSTEM||'/'||SUBSTR(SUBSTR(JOB_NAME,8),POSSTR(SUBSTR(JOB_NAME,8),'/')+1) ELSE RIGHT(JOB_NAME, LENGTH(JOB_NAME)-LOCATE('/',JOB_NAME)) END AS JOB_NAME,SUM(DOUBLE(CPU_TIME)) AS CPUTIME,SUM(DOUBLE(TOTAL_DISK_IO_COUNT)) AS TOTAL_DISK_IO_COUNT FROM TABLE(QSYS2.ACTIVE_JOB_INFO('NO')) X WHERE RIGHT(JOB_NAME, LENGTH(JOB_NAME)-LOCATE('/',JOB_NAME, LOCATE('/', JOB_NAME)+1)) IN ({','.join(self.jobsFilter)}) GROUP BY CASE WHEN SUBSYSTEM IS NOT NULL THEN SUBSTR(JOB_NAME,8,POSSTR(SUBSTR(JOB_NAME,8),'/')-1)||'/'||SUBSYSTEM||'/'||SUBSTR(SUBSTR(JOB_NAME,8),POSSTR(SUBSTR(JOB_NAME,8),'/')+1) ELSE RIGHT(JOB_NAME, LENGTH(JOB_NAME)-LOCATE('/',JOB_NAME)) END FETCH FIRST {MAX_RESULTS} ROWS ONLY", 'splitName': 'Job', 'splitBy': 'JOB_NAME', 'metricType': 'absolute'})
        
        self.queries.append({'name': 'systemStatus2', 'metricMap': [{'sqlField': 'ACTIVE_THREADS_IN_SYSTEM', 'metricName':'System.ActiveThreads', 'owner': 'System'}, {'sqlField': 'CURRENT_TEMPORARY_STORAGE', 'metricName':'System.DiskTemporaryUsed', 'owner': 'System'}, {'sqlField': 'SYSTEM_ASP_USED', 'metricName':'System.DiskSystemASPUsed', 'owner': 'System'}, {'displayField': 'Disk Total Auxiliary Storage', 'sqlField': 'TOTAL_AUXILIARY_STORAGE', 'owner': 'property', 'format':'true', 'unit': 'MB'}, {'displayField': 'Disk ASP Storage', 'sqlField': 'SYSTEM_ASP_STORAGE', 'owner': 'property', 'format':'true', 'unit': 'MB'}, {'displayField': 'CPU Capacity', 'sqlField': 'CURRENT_CPU_CAPACITY', 'owner': 'property', 'format':'true'}, {'displayField': 'Configured CPUs', 'sqlField': 'CONFIGURED_CPUS', 'owner': 'property'}, {'sqlField': 'ACTIVE_JOBS_IN_SYSTEM', 'metricName':'System.JobsActive', 'owner': 'System'}, {'displayField': 'Maximum jobs', 'sqlField': 'MAXIMUM_JOBS_IN_SYSTEM', 'owner': 'property', 'format':'true'}, {'sqlField': 'TOTAL_JOBS_IN_SYSTEM', 'metricName':'System.JobsTotal', 'owner': 'System'}], 'sql': 'SELECT ACTIVE_THREADS_IN_SYSTEM,CURRENT_TEMPORARY_STORAGE,SYSTEM_ASP_USED,TOTAL_AUXILIARY_STORAGE,SYSTEM_ASP_STORAGE,CURRENT_CPU_CAPACITY,CONFIGURED_CPUS,ACTIVE_JOBS_IN_SYSTEM,MAXIMUM_JOBS_IN_SYSTEM,TOTAL_JOBS_IN_SYSTEM FROM QSYS2.SYSTEM_STATUS_INFO', 'metricType': 'absolute'})
        self.queries.append({'name': 'envStatus', 'metricMap': [{'displayField': 'Total CPUs', 'sqlField': 'TOTAL_MEMORY', 'owner': 'property', 'format':'true', 'unit': 'MB'}, {'displayField': 'Configured Memory', 'sqlField': 'CONFIGURED_MEMORY', 'owner': 'property', 'format':'true', 'unit': 'MB'}, {'displayField': 'OS Version', 'sqlField': "OS_NAME CONCAT ' ' CONCAT OS_VERSION CONCAT '.' CONCAT OS_RELEASE", 'sqlFieldAlias': 'OS_VERSION', 'owner': 'property'}, {'displayField': 'Total CPUs', 'sqlField': 'TOTAL_CPUS', 'owner': 'property'}], 'sql': "SELECT TOTAL_MEMORY,CONFIGURED_MEMORY,OS_NAME || ' ' || OS_VERSION || '.' || OS_RELEASE AS OS_VERSION,TOTAL_CPUS FROM SYSIBMADM.ENV_SYS_INFO", 'metricType': 'absolute'})
        if self.config['job_queues'] != '':
            #self.queries.append({'name': 'queueStatus', 'metricMap': [{'sqlField': 'TOTAL_MESSAGES', 'metricName':'Queue.TotalMessages', 'owner': 'Queue'}, {'sqlField': 'MESSAGE_QUEUE_NAME', 'owner': 'Queue'}], 'sql': f'SELECT COUNT(MESSAGE_ID) AS TOTAL_MESSAGES,MESSAGE_QUEUE_NAME FROM QSYS2.MESSAGE_QUEUE_INFO {self.queuesFilterSql} GROUP BY MESSAGE_QUEUE_NAME FETCH FIRST {MAX_RESULTS} ROWS ONLY', 'splitName': 'Queue', 'splitBy': 'MESSAGE_QUEUE_NAME', 'metricType': 'absolute'})
            #self.queries.append({'name': 'jobQueueStatus', 'metricMap': [{'sqlField': 'NUMBER_OF_JOBS', 'metricName':'JobQueue.TotalJobs', 'owner': 'JobQueue'}, {'sqlField': 'ACTIVE_JOBS', 'metricName':'JobQueue.ActiveJobs', 'owner': 'JobQueue'}, {'sqlField': 'HELD_JOBS', 'metricName':'JobQueue.HeldJobs', 'owner': 'JobQueue'}, {'sqlField': 'RELEASED_JOBS', 'metricName':'JobQueue.ReleasedJobs', 'owner': 'JobQueue'}, {'sqlField': 'SCHEDULED_JOBS', 'metricName':'JobQueue.ScheduledJobs', 'owner': 'JobQueue'}, {'sqlField': 'JOB_QUEUE_NAME', 'owner': 'JobQueue'}, {'sqlField': 'JOB_QUEUE_LIBRARY', 'owner': 'JobQueue'}], 'sql': f'SELECT SUM(NUMBER_OF_JOBS) AS NUMBER_OF_JOBS,JOB_QUEUE_NAME, JOB_QUEUE_LIBRARY, SUM(ACTIVE_JOBS) AS ACTIVE_JOBS, SUM(HELD_JOBS) AS HELD_JOBS, SUM(RELEASED_JOBS) AS RELEASED_JOBS, SUM(SCHEDULED_JOBS) AS SCHEDULED_JOBS FROM QSYS2.JOB_QUEUE_INFO {self.jobQueuesFilterSql} GROUP BY JOB_QUEUE_NAME, JOB_QUEUE_LIBRARY FETCH FIRST {MAX_RESULTS} ROWS ONLY', 'splitName': 'JobQueue', 'splitBy': 'JOB_QUEUE_NAME', 'metricType': 'absolute'})
            self.queries.append({'name': 'jobQueueStatus', 'category': 'Job queue', 'metricMap': [{'sqlField': 'NUMBER_OF_JOBS', 'metricName':'JobQueue.TotalJobs', 'owner': 'JobQueue'}, {'sqlField': 'ACTIVE_JOBS', 'metricName':'JobQueue.ActiveJobs', 'owner': 'JobQueue'}, {'sqlField': 'HELD_JOBS', 'metricName':'JobQueue.HeldJobs', 'owner': 'JobQueue'}, {'sqlField': 'RELEASED_JOBS', 'metricName':'JobQueue.ReleasedJobs', 'owner': 'JobQueue'}, {'sqlField': 'SCHEDULED_JOBS', 'metricName':'JobQueue.ScheduledJobs', 'owner': 'JobQueue'}, {'sqlField': 'JOB_QUEUE_NAME', 'owner': 'JobQueue'}], 'sql': f'SELECT SUM(NUMBER_OF_JOBS) AS NUMBER_OF_JOBS,JOB_QUEUE_NAME, SUM(ACTIVE_JOBS) AS ACTIVE_JOBS, SUM(HELD_JOBS) AS HELD_JOBS, SUM(RELEASED_JOBS) AS RELEASED_JOBS, SUM(SCHEDULED_JOBS) AS SCHEDULED_JOBS FROM QSYS2.JOB_QUEUE_INFO {self.jobQueuesFilterSql} GROUP BY JOB_QUEUE_NAME FETCH FIRST {MAX_RESULTS} ROWS ONLY', 'splitName': 'JobQueue', 'splitBy': 'JOB_QUEUE_NAME', 'metricType': 'absolute'})
        if self.config['message_queues'] != '':
            self.queries.append({'name': 'messageQueueStatus', 'category': 'Message queue', 'metricMap': [{'sqlField': 'TOTAL_MESSAGES', 'metricName':'MessageQueue.TotalMessages', 'owner': 'MessageQueue'}, {'sqlField': 'MESSAGE_QUEUE_NAME', 'owner': 'MessageQueue'}], 'sql': f'SELECT COUNT(MESSAGE_ID) AS TOTAL_MESSAGES,MESSAGE_QUEUE_NAME FROM QSYS2.MESSAGE_QUEUE_INFO {self.messageQueuesFilterSql} GROUP BY MESSAGE_QUEUE_NAME FETCH FIRST {MAX_RESULTS} ROWS ONLY', 'splitName': 'MessageQueue', 'splitBy': 'MESSAGE_QUEUE_NAME', 'metricType': 'absolute'})
        
        if self.version >= 7.2:
            # Add SQL CPU Utilization if version is greater than 7.2
            for idx, item in enumerate(self.queries):
                if 'systemStatus2' in item['name']:
                    self.queries[idx]['metricMap'].append({'sqlField': 'SQL_CPU_UTILIZATION', 'metricName':'System.SqlCpu', 'owner': 'System'})
                    self.queries[idx]['sql'] = 'SELECT ACTIVE_THREADS_IN_SYSTEM,CURRENT_TEMPORARY_STORAGE,SYSTEM_ASP_USED,TOTAL_AUXILIARY_STORAGE,SYSTEM_ASP_STORAGE,CURRENT_CPU_CAPACITY,CONFIGURED_CPUS,ACTIVE_JOBS_IN_SYSTEM,MAXIMUM_JOBS_IN_SYSTEM,TOTAL_JOBS_IN_SYSTEM,SQL_CPU_UTILIZATION FROM QSYS2.SYSTEM_STATUS_INFO'
                    if (self.version == 7.3 and self.tr_level >= 9) or (self.version == 7.4 and self.tr_level >= 3): # IBM added a new view for better performance but only on 7.3+
                        self.queries[idx]['sql'] = self.queries[idx]['sql'].replace('QSYS2.SYSTEM_STATUS_INFO', 'QSYS2.SYSTEM_STATUS_INFO_BASIC')
                elif 'systemStatus1' in item['name'] and ((self.version == 7.3 and self.tr_level >= 9) or (self.version == 7.4 and self.tr_level >= 3)): # IBM added a new view for better performance but only on 7.3TR9+
                    self.queries[idx]['sql'] = self.queries[idx]['sql'].replace('QSYS2.SYSTEM_STATUS_INFO', 'QSYS2.SYSTEM_STATUS_INFO_BASIC')
            if 'Catalog' in self.config['perf_data_source']:
                self.queries.append({'name': 'diskStats', 'metricMap': [{'sqlField': 'DISKCAPACITY', 'metricName':'Disk.Capacity', 'owner': 'Disk'}, {'sqlField': 'DISKUSED', 'metricName':'Disk.Used', 'owner': 'Disk'}], 'sql': f"SELECT 'DISK ' CONCAT UNIT_NUMBER AS DISKUNITNAME,DOUBLE(UNIT_STORAGE_CAPACITY/1024)/1024 AS DISKCAPACITY,DOUBLE(UNIT_STORAGE_CAPACITY-UNIT_SPACE_AVAILABLE)/1024/1024 AS DISKUSED FROM QSYS2.SYSDISKSTAT WHERE UNIT_STORAGE_CAPACITY > 0 FETCH FIRST {self.max_disks} ROWS ONLY", 'splitName': 'Disk', 'splitBy': 'DISKUNITNAME', 'metricType': 'absolute'})
            